/**
 * Copyright Marcus Britanicus <marcusbritanicus@gmail.com>
 *
 * This file is a part of LibDesQ (https://gitlab.com/DesQ/libdesq)
 * This library contains various  Core and Gui classes which are used
 * through out the DesQ Project.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * at your option, any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301, USA.
 **/

#include <signal.h>

#include "desq/CoreApplication.hpp"
#include "desq-config.h"

#include "desq/IpcServer.hpp"
#include "desq/IpcClient.hpp"

#include "desq/Xdg.hpp"

static inline QString getSocketName( QString appId ) {
    QString sockName;

    /* $__DESQ_WORK_DIR is created and set by the session */
    if ( qgetenv( "__DESQ_WORK_DIR" ).count() ) {
        sockName = QString( qgetenv( "__DESQ_WORK_DIR" ) ) + "/" + appId;
    }

    /* It does not exist: Let's create it */
    else {
        /* We're going to assume that XDG_RUNTIME_DIR exists and is writable at this point */
        QString XDG_RUNTIME_DIR = DesQ::XDG::xdgRuntimeDir();

        /* We're also going to assume that XDG_SESSION_ID exists */
        QString XDG_SESSION_ID( qgetenv( "XDG_SESSION_ID" ) );

        /* Create this path blindly. This MUST work: No fallbacks */
        QDir( XDG_RUNTIME_DIR ).mkpath( "DesQSession-" + XDG_SESSION_ID + "/" );

        /* Our socket lives here */
        sockName = QDir( XDG_RUNTIME_DIR ).filePath( "DesQSession-" + XDG_SESSION_ID + "/" + appId );
    }

    return sockName;
}


DesQ::CoreApplication::CoreApplication( const QString& appId, int& argc, char **argv ) : QCoreApplication( argc, argv ) {
    /* App ID */
    mAppId = appId;

    /* Socket Name */
    mSocketName = getSocketName( appId );

    /* Lock File */
    lockFile = new QLockFile( mSocketName + ".lock" );

    /* Try to lock the @lockFile, if it fails, then we're not the first instance */
    if ( lockFile->tryLock() ) {
        /* IPC Server for communication */
        mIpcServer = new DesQ::IPC::Server( this );

        /* Start the server */
        bool res = mIpcServer->listen( mSocketName );

        /*
         * @res can't be false at the moment, because we're the first instance.
         * The only reason why @res is false, the socket file exists from a previous
         * crash. So delete it and try again.
         */
        if ( not res && (mIpcServer->serverError() == QAbstractSocket::AddressInUseError) ) {
            QLocalServer::removeServer( mSocketName );
            res = mIpcServer->listen( mSocketName );

            if ( !res ) {
                qWarning( "DesQ::CoreApplication: listen on local socket failed, %s", qPrintable( mIpcServer->errorString() ) );
            }
        }

        /* Irrespective of what happens, we will try to connect newConnection to receiveConnection */
        QObject::connect( mIpcServer, &DesQ::IPC::Server::messageReceived, this, &DesQ::CoreApplication::messageReceived );
    }

    /** Handle signals
     *  SIGKILL and SIGSTOP cannot be caught, blocked or ignored.
     *  SIGINT, SIGTERM, SIGABRT, SIGQUIT, SIGSEGV can all be handled.
     */
    signal( SIGINT,  DesQ::CoreApplication::signalHandler );
    signal( SIGTERM, DesQ::CoreApplication::signalHandler );
    signal( SIGQUIT, DesQ::CoreApplication::signalHandler );
    signal( SIGABRT, DesQ::CoreApplication::signalHandler );
    signal( SIGSEGV, DesQ::CoreApplication::signalHandler );
}


DesQ::CoreApplication::~CoreApplication() {
    qApp->disconnect();

    if ( mIpcServer ) {
        mIpcServer->deleteLater();
    }

    delete lockFile;
}


bool DesQ::CoreApplication::isRunning() {
    /*
     * If we have the lock, we're the server
     * In other words, if we're not there, there is no server
     */
    if ( lockFile->isLocked() ) {
        return false;
    }

    /* If we cannot get the lock then the server is running elsewhere */
    if ( not lockFile->tryLock() ) {
        return true;
    }

    /* By default, we'll assume that the server is running elsewhere */
    return true;
}


QString DesQ::CoreApplication::id() const {
    return mAppId;
}


DesQ::IPC::Server * DesQ::CoreApplication::ipcServer() {
    return mIpcServer;
}


bool DesQ::CoreApplication::sendMessage( const QString& message ) {
    if ( not isRunning() ) {
        return false;
    }

    /* Preparing socket */
    DesQ::IPC::Client socket( DesQ::IPC::Client::Message, this );

    /* Connecting to server */
    socket.connectToServer( mSocketName );

    /* Wait for ACK */
    if ( not socket.waitForAck() ) {
        return false;
    }

    /* Send the message to the server */
    return socket.sendMessage( message.toUtf8() );
}


void DesQ::CoreApplication::disconnect() {
    if ( mIpcServer ) {
        mIpcServer->close();
    }

    lockFile->unlock();
}


void DesQ::CoreApplication::signalHandler( int signum ) {
    /** Some Core Settings applicable across all DesQ Executables */
    QSettings *coreSett = new QSettings( ConfigPath + qApp->mAppId + ".conf", QSettings::IniFormat );

    switch ( signum ) {
        /** Clean up the temporary files and cease activities immediately. */
        case SIGINT: {
            if ( coreSett->value( "Core/HandleSigInt", true ).toBool() ) {
                qApp->disconnect();
                qApp->quit();

                return;
            }

            else {
                /** No more contact with the external world. */
                qApp->disconnect();

                qInfo() << "The current process has been interrupted. Cleaning up.";
                qInfo() << "If you need to exit immediately, please send SIGABRT or SIGKILL to kill the process.";
                emit qApp->interrupted();
                break;
            }
        }

        /** Ask the user if they want to save the data, then clean up the state */
        case SIGTERM: {
            if ( coreSett->value( "Core/HandleSigTerm", true ).toBool() ) {
                qApp->disconnect();
                qApp->quit();

                return;
            }
        }

        case SIGQUIT: {
            if ( coreSett->value( "Core/HandleSigQuit", true ).toBool() ) {
                qApp->disconnect();
                qApp->quit();

                return;
            }

            else {
                /** No more contact with the external world. */
                qApp->disconnect();

                /**
                 * Politely inform the user that the app is being closed.
                 * Can we request wayfire to ping the app for us to know if it's alive?
                 */

                qInfo() << "Requesting the app to gracefully shutdown. Please wait for user confirmation.";
                qInfo() << "If you need to exit immediately, please send SIGABRT or SIGKILL to kill the process.";
                emit qApp->beginShutdown();
                break;
            }
        }

        case SIGABRT: {
            if ( coreSett->value( "Core/HandleSigAbrt", true ).toBool() ) {
                qApp->disconnect();
                qApp->quit();

                return;
            }

            else {
                /** No more contact with the external world. */
                qApp->disconnect();

                qInfo() << "Try to save state and abort.";
                emit qApp->aborted();
                break;
            }
        }

        case SIGSEGV: {
            if ( coreSett->value( "Core/HandleSigSegv", true ).toBool() ) {
                qApp->disconnect();
                qApp->quit();

                return;
            }

            else {
                /** No more contact with the external world. */
                qApp->disconnect();

                qInfo() << "Try to save state and abort.";
                emit qApp->aborted();
                break;
            }
        }

        default: {
            qDebug() << "Not handing this signal:" << signum;
            break;
        }
    }
}
